#!C:/Programs/Strawberry/win64/perl/bin/perl.exe -I"."

#####################################
#             Chemistry             #
#####################################
# Quantitative analysis of proteins #
#####################################

use strict; use warnings;

use CGI;
use CGI::Carp "fatalsToBrowser";
use File::Basename;
$CGI::POST_MAX = 1024 * 25;
my $cgi = new CGI;
print "Content-Type: text/html\n\n";
print '<!DOCTYPE html>', "\n\n";
my %amino_acids = (
  'A' => { 'code3' => 'Ala', 'name' => 'Alanine' },
  'C' => { 'code3' => 'Cys', 'name' => 'Cysteine' },
  'D' => { 'code3' => 'Asp', 'name' => 'Aspartic acid' },
  'E' => { 'code3' => 'Glu', 'name' => 'Glutamic acid' },
  'F' => { 'code3' => 'Phe', 'name' => 'Phenylalanine' },
  'G' => { 'code3' => 'Gly', 'name' => 'Glycine' },
  'H' => { 'code3' => 'His', 'name' => 'Histidine' },
  'I' => { 'code3' => 'Ile', 'name' => 'Isoleucine' },
  'K' => { 'code3' => 'Lys', 'name' => 'Lysine' },
  'L' => { 'code3' => 'Leu', 'name' => 'Leucine' },
  'M' => { 'code3' => 'Met', 'name' => 'Methionine' },
  'N' => { 'code3' => 'Asn', 'name' => 'Asparagine' },
  'P' => { 'code3' => 'Pro', 'name' => 'Proline' },
  'Q' => { 'code3' => 'Gln', 'name' => 'Glutamine' },
  'R' => { 'code3' => 'Arg', 'name' => 'Arginine' },
  'S' => { 'code3' => 'Ser', 'name' => 'Serine' },
  'T' => { 'code3' => 'Thr', 'name' => 'Threonine' },
  'V' => { 'code3' => 'Val', 'name' => 'Valine' },
  'W' => { 'code3' => 'Trp', 'name' => 'Tryptophan' },
  'Y' => { 'code3' => 'Tyr', 'name' => 'Tyrosine' }
);
my %aa_counts;
foreach my $aa (keys %amino_acids) {
  $aa_counts{$amino_acids{$aa}{'name'}} = 0;
};
my @categories = ('Molecule size', 'Side chain', 'Polarity', 'Charge', 'Hydropathy', 'Requirement', 'Metabolism');
my @size_counts; my @chain_counts; my @polarity_counts; my @charge_counts;
my @hydro_counts; my @requirement_counts; my @function_counts;
my $app_dir = "C:\\Programs\\Apache24\\htdocs\\computing\\website\\applications";
my $filename = ''; my $fileusage = ''; my $mess = ''; my $seq;
my $protein; my $protein_length = 0;
# Get parameters from 'calling' webpage
my %params = $cgi->Vars;
my $action = lc($params{'action'});
if ($action eq 'calculate') {
  # Get protein sequence (from file or from text area)
  $fileusage = $params{'usefile'}; $filename = $params{'filename'};
  if ($fileusage eq 'usefile') {
    # Get protein sequence from file
    if ($filename) {
      my ( $name, $path, $extension ) = fileparse ( $filename, '..*' );
      $filename = $name . $extension; my $filename2 = $filename;
      $filename2 =~ s/[^a-zA-Z0-9 _\.\-]//g;
      if ($filename2 eq $filename) {
        my $FH = $cgi->upload("filename");
        while (<$FH>) {
          $seq .= $_;
        }
      }
      else {
        $mess = "Invalid filename: $filename!";
      }
    }
    else {
      $mess = 'No filename or could not upload local protein file!';
    }
  }
  else {
    # Get protein sequence from text area
    $seq = $params{'protein'};
  }
  if ($seq) {
    # If there is a sequence, check if it is a valid protein; if yes, do calculations
    $protein = $seq;
    if (substr($protein, 0, 1) eq '>') {
      # Remove FASTA header (if there is any)
      $protein =~ /\n/g;
      my $p = pos($protein);
      $protein = substr($seq, $p, length($seq) - $p);
    }
    $protein =~ s/\s//g;
    $protein_length = length($protein);
    if (valid_protein($protein)) {
      # If the sequence is a valid protein, do the calculations
      my ($ref_aa_counts, $ref_size_counts, $ref_chain_counts, $ref_polarity_counts, $ref_charge_counts,
        $ref_hydro_counts, $ref_requirement_counts, $ref_function_counts) = analyze($protein, \%amino_acids, \%aa_counts);
      %aa_counts =  %$ref_aa_counts; @size_counts = @$ref_size_counts; @chain_counts = @$ref_chain_counts;
      @polarity_counts = @$ref_polarity_counts; @charge_counts = @$ref_charge_counts; @hydro_counts = @$ref_hydro_counts;
        @requirement_counts = @$ref_requirement_counts; @function_counts = @$ref_function_counts;
    }
    else {
      # If the sequence isn't a valid protein, display error message
      unless ($mess) {
        $mess = 'Invalid protein sequence!';
      }
    }
  }
  else {
    # If there isn't a sequence, display error message
    unless ($mess) {
      $mess = 'No protein sequence...';
    }
  }
}
# Read template HTML file
my $template = "$app_dir\\proteins.template.html";
open(my $input, "<", $template)
  or die "Can't open template file $template: $!";
my @lines = <$input>;
close($input);
# Create the webpage from template HTML file
# Replace all custom tags by actual values
foreach my $line (@lines) {
  chomp($line);
  if ($line) {
    if (substr($line, 0, 1) eq '#') {                                          # a '#' at beginning of line indicates that line contains a tag
      $line = substr($line, 1);                                                # remove the '#' at beginning of line
      if ($line =~ /#len#/) {
        if ($action eq 'calculate' and !$mess) {
          my $len = ' (' . $protein_length . ' amino acids)';
          $line =~ s/#len#/$len/;                                              # fill in the protein sequence length (or nothing)
        }
        else {
          $line =~ s/#len#//;
        }
      }
      if ($line =~ /#protein#/) {
        $line =~ s/#protein#/$seq/;                                            # fill in the protein sequence (empty if there isn't any)
      }
      elsif ($line =~ /#checked_usefile#/) {
        if ($fileusage eq 'usefile') {
          $line =~ s/#checked_usefile#/checked="checked"/;                     # check 'load protein from file' checkbox
        }
        else {
          $line =~ s/#checked_usefile#//;
        }
      }
      elsif ($line =~ /#counts#/) {
        # Create HTML table (actually 2 partial tables) with amino acids counts and percentages
        if ($action eq 'calculate' and !$mess) {
          my $table;
          foreach my $aa (sort keys %aa_counts) {
            if ($aa eq 'Alanine' or $aa eq 'Leucine') {
              # These are the first items of the 2 partial tables
              if ($aa eq 'Alanine') {
                # This is the very first item (create outer table)
                $table = '<table border="0"><tr>';
              }
              # Create inner table (the one actually displayed) and table header row
              $table .= '<td><table border="1" style="margin-top:25px">';
              $table .= '<tr><th>Amino acid</th><th>Count</th><th>Percent</th></tr>';
            }
            # Create table data rows
            $table .= '<tr><td style="padding-right:10px">' . $aa . '</td>';
            $table .= '<td  style="text-align: right">' . $aa_counts{$aa} . '</td>';
            $table .= '<td  style="text-align: right">' . sprintf("%.2f", 100 * $aa_counts{$aa} / $protein_length) . '%</td></tr>';
            if ($aa eq 'Isoleucine') {
              # Last item of first partial table (end of table HTML)
              $table .= '</table></td><td width="38px">&nbsp;</td>';
            }
          }
          # End of partial and outer table HTML
          $table .= '</table></td></tr></table>';
          # Replace the "#counts#" tag by generated the table HTML
          $line =~ s/#counts#/$table/;
        }
        else {
          # If the call to the program is sth else than for a calculation, or if there are some errors,
          # replace the "#counts#" tag with an empty string
          $line =~ s/#counts#/&nbsp;/;
        }
      }
      elsif ($line =~ /#classes#/) {
        # Create the categories table links HTML
        if ($action eq 'calculate' and !$mess) {
          my $links = '';
          $links  = '<table style="margin-bottom:15px"><tr><td><a href="javascript:onoff(' . "'i_1'" . ')">' . $categories[0] . '</a></td>';
          $links .= '<td width="40px">&nbsp;</td>';
          $links .= '<td><a href="javascript:onoff(' . "'i_2'" . ')">' . $categories[1] . '</a></td>';
          $links .= '<td width="40px">&nbsp;</td>';
          $links .= '<td><a href="javascript:onoff(' . "'i_3'" . ')">' . $categories[2] . '</a></td>';
          $links .= '<td width="40px">&nbsp;</td>';
          $links .= '<td><a href="javascript:onoff(' . "'i_4'" . ')">' . $categories[3] . '</a></td>';
          $links .= '</tr>';
          $links .= '<tr><td><a href="javascript:onoff(' . "'i_5'" . ')">' . $categories[4] . '</a></td>';
          $links .= '<td width="40px">&nbsp;</td>';
          $links .= '<td><a href="javascript:onoff(' . "'i_6'" . ')">' . $categories[5] . '</a></td>';
          $links .= '<td width="40px">&nbsp;</td>';
          $links .= '<td><a href="javascript:onoff(' . "'i_7'" . ')">' . $categories[6] . '</a></td>';
          $links .= '</tr></table>';
          # Replace the "#classes#" tag with the generated links HTML
          $line =~ s/#classes#/$links/;
        }
        else {
          # If the call to the program is sth else than for a calculation, or if there are some errors,
          # replace the "#classes#" tag with an empty string
          $line =~ s/#classes#/&nbsp;/;
        }
      }
      elsif ($line =~ /#category/) {
        # In the 7 categories tables, replace the "#category0#" - "#category6#" tags by the 7 category names (Molecule size, Side chain, etc)
        if ($action eq 'calculate' and !$mess) {
          for (my $i = 0; $ i <= 6; $i++) {
            my $tag_category = '#category' . $i . '#';
            my $category = $categories[$i];
            $line =~ s/$tag_category/$category/;
          }
        }
      }
      elsif ($line =~ /#size/) {
        # In the "size" category table, replace the tags by the counts resp. percentages for the 5 size items
        if ($action eq 'calculate' and !$mess) {
          for (my $i = 0; $ i <= 4; $i++) {
            my $tag_size = '#size' . $i . '#';
            my $tag_psize = '#sizep' . $i . '#';
            my $sizes = $size_counts[$i];
            my $psizes = sprintf("%.2f", 100 * $sizes / $protein_length) . '%';
            $line =~ s/$tag_size/$sizes/;
            $line =~ s/$tag_psize/$psizes/;
          }
        }
      }
      elsif ($line =~ /#chain/) {
        # In the "sidechain" category table, replace the tags by the counts resp. percentages for the 7 sidechain items
        if ($action eq 'calculate' and !$mess) {
          for (my $i = 0; $ i <= 6; $i++) {
            my $tag_chain = '#chain' . $i . '#';
            my $tag_pchain = '#chainp' . $i . '#';
            my $chains = $chain_counts[$i];
            my $pchains = sprintf("%.2f", 100 * $chains / $protein_length) . '%';
            $line =~ s/$tag_chain/$chains/;
            $line =~ s/$tag_pchain/$pchains/;
          }
        }
      }
      elsif ($line =~ /#polar/) {
        # In the "polarity" category table, replace the tags by the counts resp. percentages for the 2 polarity items
        if ($action eq 'calculate' and !$mess) {
          for (my $i = 0; $ i <= 1; $i++) {
            my $tag_polar = '#polar' . $i . '#';
            my $tag_ppolar = '#polarp' . $i . '#';
            my $polars = $polarity_counts[$i];
            my $ppolars = sprintf("%.2f", 100 * $polars / $protein_length) . '%';
            $line =~ s/$tag_polar/$polars/;
            $line =~ s/$tag_ppolar/$ppolars/;
          }
        }
      }
      elsif ($line =~ /#charge/) {
        # In the "charge" category table, replace the tags by the counts resp. percentages for the 3 charge items
        if ($action eq 'calculate' and !$mess) {
          for (my $i = 0; $ i <= 2; $i++) {
            my $tag_charge = '#charge' . $i . '#';
            my $tag_pcharge = '#chargep' . $i . '#';
            my $charges = $charge_counts[$i];
            my $pcharges = sprintf("%.2f", 100 * $charges / $protein_length) . '%';
            $line =~ s/$tag_charge/$charges/;
            $line =~ s/$tag_pcharge/$pcharges/;
          }
        }
      }
      elsif ($line =~ /#hydro/) {
        # In the "hydropathy" category table, replace the tags by the counts resp. percentages for the 3 hydropathy items
        if ($action eq 'calculate' and !$mess) {
          for (my $i = 0; $ i <= 2; $i++) {
            my $tag_hydro = '#hydro' . $i . '#';
            my $tag_phydro = '#hydrop' . $i . '#';
            my $hydros = $hydro_counts[$i];
            my $phydros = sprintf("%.2f", 100 * $hydros / $protein_length) . '%';
            $line =~ s/$tag_hydro/$hydros/;
            $line =~ s/$tag_phydro/$phydros/;
          }
        }
      }
      elsif ($line =~ /#nutr/) {
        # In the "requirement" category table, replace the tags by the counts resp. percentages for the 3 requirements items
        if ($action eq 'calculate' and !$mess) {
          for (my $i = 0; $ i <= 2; $i++) {
            my $tag_req = '#nutr' . $i . '#';
            my $tag_preq = '#nutrp' . $i . '#';
            my $reqs = $requirement_counts[$i];
            my $preqs = sprintf("%.2f", 100 * $reqs / $protein_length) . '%';
            $line =~ s/$tag_req/$reqs/;
            $line =~ s/$tag_preq/$preqs/;
          }
        }
      }
      elsif ($line =~ /#func/) {
        # In the "metabolism" category table, replace the tags by the counts resp. percentages for the 3 metabolism items
        if ($action eq 'calculate' and !$mess) {
          for (my $i = 0; $ i <= 2; $i++) {
            my $tag_func = '#func' . $i . '#';
            my $tag_pfunc = '#funcp' . $i . '#';
            my $funcs = $function_counts[$i];
            my $pfuncs = sprintf("%.2f", 100 * $funcs / $protein_length) . '%';
            $line =~ s/$tag_func/$funcs/;
            $line =~ s/$tag_pfunc/$pfuncs/;
          }
        }
      }
      elsif ($line =~ /#mess#/) {
        $mess = '<span class="red">' . $mess . '</span>';
        $line =~ s/#mess#/$mess/;                                              # fill in error message (or nothing, if all OK)
      }
    }
  }
  # Print file-line (with tags replaced) to webpage
  unless ($line =~ /#/) {
    print "$line\n";
  }
}

#----------------------------------#
# Check protein sequence validity  #
#----------------------------------#

sub valid_protein {
  my ($protein) = @_; my $valid = 0;
  my $aa = 'ACDEFGHIKLMNPQRSTVWY';
  if ($protein =~ /^([$aa]+)$/) {
    $valid = 1;
  }
  return $valid;
}

#--------------------------------#
# Quantitative protein analysis  #
#--------------------------------#

sub analyze {
  my ($protein, $ref_amino_acids, $ref_aa_counts) = @_;
  my %amino_acids = %$ref_amino_acids; my %aa_counts = %$ref_aa_counts;
  my @size_counts = (0, 0, 0, 0, 0); my @chain_counts = (0, 0, 0, 0, 0, 0, 0); my @polarity_counts = (0, 0);
  my @charge_counts = (0, 0, 0); my @hydro_counts = (0, 0, 0); my @requirement_counts = (0, 0, 0); my @function_counts = (0, 0, 0);
  my @sizes = ( 'AGS', 'CDNPT', 'EHQV', 'IKLMR', 'FWY' );
  my @chains = ( 'AIGLV', 'FHWY', 'P', 'CM', 'DE', 'KR', 'NQST' );
  my  @polarities = ( 'DEHKNQRSTY', 'ACFGILMPVW' );
  my  @charges = ( 'HKR', 'DE', 'ACFGILMNPSTQVWY' );
  my @hydros = ('ACFILMVW', 'DEKNQR', 'GHPSTY');
  my @requirements = ('FIKLMTVW', 'HR', 'ACDEGNPNSY');
  my @functions = ('KL', 'FITWY', 'ACDEGHMNPQRSV');
  for (my $i = 0; $i < length($protein); $i++) {
    my $aa = substr($protein, $i, 1);
    # Amino acid counts
    $aa_counts{$amino_acids{$aa}{'name'}}++;
    for (my $j = 0; $j <= 4; $j++) {
      # Size category counts
      if ($aa =~ /[$sizes[$j]]/) {
        $size_counts[$j]++;
      }
    }
    for (my $j = 0; $j <= 6; $j++) {
      # Sidechain category counts
      if ($aa =~ /[$chains[$j]]/) {
        $chain_counts[$j]++;
      }
    }
    for (my $j = 0; $j <= 1; $j++) {
      # Polarity category counts
      if ($aa =~ /[$polarities[$j]]/) {
        $polarity_counts[$j]++;
      }
    }
    for (my $j = 0; $j <= 2; $j++) {
      # Charge category counts
      if ($aa =~ /[$charges[$j]]/) {
        $charge_counts[$j]++;
      }
      # Hydropathy category counts
      if ($aa =~ /[$hydros[$j]]/) {
        $hydro_counts[$j]++;
      }
      # Requirement category counts
      if ($aa =~ /[$requirements[$j]]/) {
        $requirement_counts[$j]++;
      }
      # Metabolism category counts
      if ($aa =~ /[$functions[$j]]/) {
        $function_counts[$j]++;
      }
    }
  }
  return (\%aa_counts, \@size_counts, \@chain_counts, \@polarity_counts, \@charge_counts, \@hydro_counts,  \@requirement_counts,  \@function_counts);
}
